import cliutils, docopt, json, logging, os, re, random, sequtils,
  times, timeutils

from posix import SIGTERM, SIGHUP, signal, kill
import strutils except toUpper
from unicode import toUpper

type
  ParseState = enum BeforeHeading, ReadingPlans, AfterPlans
  PlanItem* = tuple[time: TimeInfo, note: string]
  DPConfig = tuple[planDir, dateFmt, pidfile, logfile, errfile: string,
    notificationSecs: int]

const VERSION = "0.3.0"
const timeFmt = "HH:mm"

var args: Table[string, Value]
var cfg: DPConfig
let appName = getAppFilename()
let soundsDir: string = appName[0..(appName.len-15)] & "sounds/navi"
let soundFiles = toSeq(walkFiles(soundsDir))
randomize()

proc parseDailyPlan(filename: string): seq[PlanItem] =

  debug "Parsing daily plan file: " & filename

  result = @[]

  var parseState = BeforeHeading
  let planItemRe = re"\s*(\d{4})\s+(.*)"
  const timeFmt = "HHmm"

  for line in lines filename:
    if line.strip.len == 0: continue
    case parseState

    of BeforeHeading:
      if line.strip.startsWith("## Timeline"): parseState = ReadingPlans

    of ReadingPlans:
      # TODO: This is the better code using the nre module: 
      # let match = line.find(planItemRe)
      # if match.isSome(): result.add((
      #   time: parse(match.get().captures[0], timeFmt),
      #   note: match.get().captures[1]))
      # else: parseState = AfterPlans
      #
      # Curently there is an incompatibility between the os and nre modules
      # (see https://github.com/nim-lang/Nim/issues/4996) so the following code
      # is used to avoid using nre until the bug is fixed.

      if line.match(planItemRe):
        let stripped = line.strip
        result.add((
          time: parse(stripped[0..4], timeFmt),
          note: stripped.substr(5)))
      else: parseState = AfterPlans

    of AfterPlans: break
    else: break

  debug "Found " & $result.len & " items."

proc notifyDailyPlanItem(item: PlanItem): void =

  let soundFile = appName[0..(appName.len-15)] & "sounds/navi/" & soundFiles[random(soundFiles.len)]
  case hostOS
  of "macosx":
    discard execShellCmd("osascript -e 'display notification \"" &
      item.time.format(timeFmt) & " - " & item.note &
      "\" with title \"Daily Notifier v" & VERSION & "\"")
    discard execShellCmd("ogg123 \"" & soundFile & "\"")

  of "linux":
    discard execShellCmd("notify-send 'Daily Notifier v" & VERSION & "' '" & item.note & "'")
    discard execShellCmd("paplay \"" & soundFile & "\"")

  else: quit("Unsupported host OS: '" & hostOS & "'.")

proc loadConfig(s: cint): void {. exportc, noconv .} =

  # Find and parse the .dailynotificationrc file
  let rcLocations = @[
    if args["--config"]: $args["<cfgFile>"] else:"",
    ".dailynotificationrc", $getEnv("DAILY_NOTIFICATION_RC"),
    $getEnv("HOME") & "/.dailynotificationrc"]

  var rcFilename: string =
    foldl(rcLocations, if len(a) > 0: a elif existsFile(b): b else: "")

  debug "Reloading config file from " & rcFilename

  var jsonCfg: JsonNode
  try: jsonCfg = parseFile(rcFilename)
  except: jsonCfg = newJObject()

  var ccfg = CombinedConfig(docopt: args, json: jsonCfg)

  cfg = (
    ccfg.getVal("plans-directory", "plans"),
    ccfg.getVal("date-format", "yyyy-MM-dd"),
    ccfg.getVal("pidfile", "/tmp/daily-plans.pid"),
    ccfg.getVal("logfile", "/tmp/daily-plans.log"),
    ccfg.getVal("errfile", "/tmp/daily-plans.error.log"),
    parseInt(ccfg.getVal("notification-seconds", "600"))
  )

  if not existsDir(cfg.planDir):
    quit("daily_notifier: plan directory does not exist: '" & cfg.planDir & "'", 1)

proc mainLoop(args: Table[string, Value]): void =
  
  loadConfig(0)
  signal(SIGHUP, loadConfig)
  var curDay: TimeInfo = getLocalTime(fromSeconds(0))
  var todaysFile = "nofile"
  var lastModTime = fromSeconds(0)
  var todaysItems: seq[PlanItem] = @[]

  while true:

    # Check the date
    let now = getLocalTime(getTime())
    let today = startOfDay(now)

    # If we need to change day, look for a new file.
    if today != curDay:
      curDay = today
      todaysFile = cfg.planDir & "/" & today.format(cfg.dateFmt) & "-plan.md"
      todaysItems = @[]
      lastModTime = fromSeconds(0)

    # Check for our plan file.
    if fileExists(todaysFile):
      let fileModTime = getLastModificationTime(todaysFile)

      # Check if the file has been modified or we have no itmes
      if fileModTime > lastModTime or todaysItems.len == 0:
        lastModTime = fileModTime
        todaysItems = parseDailyPlan(todaysFile)

    # Check to see if any items are happening soon.
    let cutoff = now + seconds(cfg.notificationSecs)
    let needsNotification = todaysItems.filterIt(it.time > now and it.time < cutoff)

    todaysItems.keepItIf(not needsNotification.contains(it))

    for item in needsNotification:
      notifyDailyPlanItem(item)

    # Sleep for a while
    sleep(30 * 1000)
  
when isMainModule:

  let doc = """
Usage:
  daily_notifier start [options]
  daily_notifier reconfigure [options]
  daily_notifier stop

Options:

  -d --plans-directory <dir>  Directory to search for plan files.

  -f --date-format <fmt> 
  
    Date pattern for identifying plan files. This is used in conjunction
    with plan directory to idenfity files that should be parsed as plan files.

  -c --config <cfgFile>       Use <cfgFile> as the source of configuration for
                              daily-notification.

  -p --pidfile <pidfile>      Daemon PID filename 
  -L --logfile <logfile>      Log file.
  -E --errfile <errfile>      Error file.
  -h --help                   Print this usage information.

  -s --notification-seconds <sec>
  -v --verbose                Enable verbose output.

    Notification period for plan items.
    
"""

  args = docopt(doc, version = "daily_notifier " & VERSION)

  if args["--help"]:
    echo doc
    quit()

  loadConfig(0)

  addHandler(newConsoleLogger(
    if args["--verbose"]: lvlAll
    else: lvlInfo))

  if args["start"]:
    # Start our daemon process (if needed)
    info "Starting daemon."
    let childPid = daemonize(cfg.pidfile, "/dev/null", cfg.logfile, cfg.errfile, proc(): void = mainLoop(args))
    if childPid == 0: quit(QuitSuccess) # We are the child... don't need to do anything else.
    info "Started, pid: " & $childPid

  elif args["stop"] or args["reconfigure"]:

    if not fileExists(cfg.pidfile):
      echo "daily_notifier is not running"
      quit(QuitSuccess)

    let pid = parseInt(readFile(cfg.pidfile).strip)
    debug "Killing process " & $pid

    if args["stop"]: discard kill(pid, SIGTERM)
    else: discard kill(pid, SIGHUP)
